@@ -38,19 +38,8 @@ use Getopt::Long;
 use File::Copy;
 use File::Basename;
 use Cwd;
-
-BEGIN {
-    # Load the URI::Escape module safely
-    eval { require URI::Escape; };
-    if ($@) {
-	my $progname = basename $0;
-	if ($@ =~ /^Can\'t locate URI\/Escape\.pm/) {
-	    die "$progname: you must have the liburi-perl package installed\nto use this script\n";
-	}
-	die "$progname: problem loading the URI::Escape module:\n  $@\nHave you installed the liburi-perl package?\n";
-    }
-    import URI::Escape;
-}
+use lib '/usr/share/devscripts';
+use Devscripts::Debbugs;
 
 # Predeclare functions
 sub fatal($);
@@ -62,6 +51,25 @@ my $modified_conf_msg;
 my %env;
 my $CHGLINE;  # used by the format O section at the end
 
+my $lpdc_broken;
+
+sub have_lpdc {
+    return ($lpdc_broken ? 0 : 1) if defined $lpdc_broken;
+    eval {
+	require Parse::DebControl;
+    };
+
+    if ($@) {
+        if ($@ =~ m%^Can\'t locate Parse/DebControl%) {
+            $lpdc_broken="the libparse-debcontrol-perl package is not installed";
+        } else {
+            $lpdc_broken="couldn't load Parse::DebControl: $@";
+        }
+    }
+    else { $lpdc_broken=''; }
+    return $lpdc_broken ? 0 : 1;
+}
+
 sub usage () {
     print <<"EOF";
 Usage: $progname [options] [changelog entry]
@@ -82,22 +90,37 @@ Options:
   --create
          Create a new changelog (default) or NEWS file (with --news) and
          open for editing
+  --empty
+         When creating a new changelog, don't add any changes to it
+         (i.e. only include the header and trailer lines)
   --package <package>
          Specify the package name when using --create (optional)
+  --auto-nmu
+         Attempt to intelligently determine whether a change to the
+         changelog represents an NMU (default)
+  --no-auto-nmu
+         Do not do so
   -n, --nmu
          Increment the Debian release number for a non-maintainer upload
-  --qa
+  --bin-nmu
+         Increment the Debian release number for a binary non-maintainer upload
+  -q, --qa
          Increment the Debian release number for a Debian QA Team upload
   -s, --security
-         Increment the Debian release number for a security update upload
+         Increment the Debian release number for a Debian Security Team upload
   -U, --upstream
          Increment the Debian release number without any distribution name
   --bpo
          Increment the Debian release number for a Backports.org upload
 	 to "etch-backports"
+  -l, --local <suffix>
+         Add a suffix to the Debian version number for a local build
   -b, --force-bad-version
          Force a version to be less than the current one (e.g., when
          backporting)
+  --allow-lower-version
+         Allow a version to be less than the current one (e.g., when
+         backporting) if it matches a specified pattern
   --force-distribution
          Force the provided distribution to be used, even if it doesn't match
          the list of known distributions
@@ -113,7 +136,7 @@ Options:
   --no-preserve
          Do not preserve the directory name (default)
   --distributor <distributor>
-  	 Override the distributor ID from lsb_release.
+         Override the distributor ID from lsb_release.
   -D, --distribution <dist>
          Use the specified distribution in the changelog entry being edited
   -u, --urgency <urgency>
@@ -150,8 +173,8 @@ Options:
          Display this help message and exit
   --version
          Display version information
-  At most one of -a, -i, -e, -r, -v, -d, -n, --qa, --bpo (or their long equivalents)
-  may be used.
+  At most one of -a, -i, -e, -r, -v, -d, -n, --bin-nmu, -q, --qa, -s, --bpo, -l
+  (or their long equivalents) may be used.
   With no options, one of -i or -a is chosen by looking for a .upload
   file in the parent directory and checking its contents.
 
@@ -180,7 +203,9 @@ my $opt_release_heuristic = 'log';
 my $opt_multimaint = 1;
 my $opt_multimaint_merge = 0;
 my $opt_tz = undef;
-my $opt_mainttrailer = 0;
+my $opt_t = '';
+my $opt_allow_lower = '';
+my $opt_auto_nmu = 'yes';
 
 # Next, read configuration files and then command line
 # The next stuff is boilerplate
@@ -199,6 +224,9 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'DEBCHANGE_MULTIMAINT' => 'yes',
 		       'DEBCHANGE_TZ' => $ENV{TZ}, # undef if TZ unset
 		       'DEBCHANGE_MULTIMAINT_MERGE' => 'no',
+		       'DEBCHANGE_MAINTTRAILER' => '',
+		       'DEBCHANGE_LOWER_VERSION_PATTERN' => '',
+		       'DEBCHANGE_AUTO_NMU' => 'yes',
 		       );
     $config_vars{'DEBCHANGE_TZ'} ||= '';
     my %config_default = %config_vars;
@@ -228,6 +256,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'DEBCHANGE_MULTIMAINT'}='yes';
     $config_vars{'DEBCHANGE_MULTIMAINT_MERGE'} =~ /^(yes|no)$/
 	or $config_vars{'DEBCHANGE_MULTIMAINT_MERGE'}='no';
+    $config_vars{'DEBCHANGE_AUTO_NMU'} =~ /^(yes|no)$/
+	or $config_vars{'DEBCHANGE_AUTO_NMU'}='yes';
 
     foreach my $var (sort keys %config_vars) {
 	if ($config_vars{$var} ne $config_default{$var}) {
@@ -245,15 +275,19 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $opt_multimaint = $config_vars{'DEBCHANGE_MULTIMAINT'} eq 'no' ? 0 : 1;
     $opt_tz = $config_vars{'DEBCHANGE_TZ'};
     $opt_multimaint_merge = $config_vars{'DEBCHANGE_MULTIMAINT_MERGE'} eq 'no' ? 0 : 1;
+    $opt_t = ($config_vars{'DEBCHANGE_MAINTTRAILER'} eq 'no' ? 0 : 1)
+	if $config_vars{'DEBCHANGE_MAINTTRAILER'};
+    $opt_allow_lower = $config_vars{'DEBCHANGE_LOWER_VERSION_PATTERN'};
+    $opt_auto_nmu = $config_vars{'DEBCHANGE_AUTO_NMU'};
 }
 
 # We use bundling so that the short option behaviour is the same as
 # with older debchange versions.
 my ($opt_help, $opt_version);
-my ($opt_i, $opt_a, $opt_e, $opt_r, $opt_v, $opt_b, $opt_d, $opt_D, $opt_u, $opt_t, $opt_force_dist);
-my ($opt_n, $opt_security, $opt_qa, $opt_U, $opt_bpo, $opt_c, $opt_m, $opt_create, $opt_package, @closes);
+my ($opt_i, $opt_a, $opt_e, $opt_r, $opt_v, $opt_b, $opt_d, $opt_D, $opt_u, $opt_force_dist);
+my ($opt_n, $opt_bn, $opt_qa, $opt_s, $opt_U, $opt_bpo, $opt_l, $opt_c, $opt_m, $opt_create, $opt_package, @closes);
 my ($opt_news);
-my ($opt_ignore, $opt_level, $opt_regex, $opt_noconf);
+my ($opt_ignore, $opt_level, $opt_regex, $opt_noconf, $opt_empty);
 my ($opt_distributor);
 
 Getopt::Long::Configure('bundling');
@@ -267,18 +301,21 @@ GetOptions("help|h" => \$opt_help,
 	   "package=s" => \$opt_package,
 	   "v|newversion=s" => \$opt_v,
 	   "b|force-bad-version" => \$opt_b,
+	   "allow-lower-version" => \$opt_allow_lower,
 	   "force-distribution" => \$opt_force_dist,
 	   "d|fromdirname" => \$opt_d,
 	   "p" => \$opt_p,
 	   "preserve!" => \$opt_p,
-	   "D|distribution=s" => \$opt_D,
 	   "distributor=s" => \$opt_distributor,
+	   "D|distribution=s" => \$opt_D,
 	   "u|urgency=s" => \$opt_u,
 	   "n|nmu" => \$opt_n,
-	   "s|security" => \$opt_security,
-	   "qa" => \$opt_qa,
+	   "bin-nmu" => \$opt_bn,
+	   "q|qa" => \$opt_qa,
+	   "s|security" => \$opt_s,
 	   "U|upstream" => \$opt_U,
 	   "bpo" => \$opt_bpo,
+	   "l|local=s" => \$opt_l,
 	   "query!" => \$opt_query,
 	   "closes=s" => \@closes,
 	   "c|changelog=s" => \$opt_c,
@@ -293,22 +330,19 @@ GetOptions("help|h" => \$opt_help,
 	   "noconf" => \$opt_noconf,
 	   "no-conf" => \$opt_noconf,
 	   "release-heuristic=s" => \$opt_release_heuristic,
+	   "empty" => \$opt_empty,
+	   "auto-nmu!" => \$opt_auto_nmu,
 	   )
     or die "Usage: $progname [options] [changelog entry]\nRun $progname --help for more details\n";
 
+# So that we can distinguish, if required, between an explicit
+# passing of -a / -i and their values being automagically deduced
+# later on
+my $opt_a_passed = $opt_a || 0;
+my $opt_i_passed = $opt_i || 0;
 $opt_news = 'debian/NEWS' if defined $opt_news and $opt_news eq '';
 
-# See if we're Debian, Ubuntu or someone else, if we can
-my $distributor;
-if (defined $opt_distributor) {
-    $distributor=$opt_distributor;
-} elsif (system('command -v lsb_release >/dev/null 2>&1') >> 8 == 0) {
-    $distributor = `lsb_release -is 2>/dev/null`;
-    chomp $distributor;
-}
-$distributor ||= 'Debian';
-
-if (! defined $opt_t && $opt_release_heuristic eq 'changelog') {
+if ($opt_t eq '' && $opt_release_heuristic eq 'changelog') {
     $opt_t = 1;
 }
 
@@ -333,25 +367,42 @@ if (defined $opt_level) {
 if (defined $opt_regex) { $check_dirname_regex = $opt_regex; }
 
 # Only allow at most one non-help option
-fatal "Only one of -a, -i, -e, -r, -v, -d, -n/--nmu, -s/--security, --qa, --bpo is allowed;\ntry $progname --help for more help"
-    if ($opt_i?1:0) + ($opt_a?1:0) + ($opt_e?1:0) + ($opt_r?1:0) + ($opt_v?1:0) + ($opt_d?1:0) + ($opt_n?1:0) + ($opt_qa?1:0) + ($opt_security?1:0) + ($opt_bpo?1:0) > 1;
+fatal "Only one of -a, -i, -e, -r, -v, -d, -n/--nmu, --bin-nmu, -q/--qa, -s/--security, --bpo, -l/--local is allowed;\ntry $progname --help for more help"
+    if ($opt_i?1:0) + ($opt_a?1:0) + ($opt_e?1:0) + ($opt_r?1:0) + ($opt_v?1:0) + ($opt_d?1:0) + ($opt_n?1:0) + ($opt_bn?1:0) + ($opt_qa?1:0) + ($opt_s?1:0) + ($opt_bpo?1:0) + ($opt_l?1:0) > 1;
+
+if ($opt_s) {
+    $opt_u = "high";
+}
 
 if (defined $opt_u) {
     fatal "Urgency can only be one of: low, medium, high, critical, emergency"
 	unless $opt_u =~ /^(low|medium|high|critical|emergency)$/;
 }
 
+# See if we're Debian, Ubuntu or someone else, if we can
+my $distributor;
+if (defined $opt_distributor) {
+    $distributor = $opt_distributor;
+} elsif (system('command -v lsb_release >/dev/null 2>&1') >> 8 == 0) {
+    $distributor = `lsb_release -is 2>/dev/null`;
+    chomp $distributor;
+}
+$distributor ||= 'Debian';
+if ($distributor eq 'Ubuntu' and ($opt_n or $opt_bn or $opt_qa or $opt_bpo)) {
+    $distributor = 'Debian';
+}
+
 # Check the distro name given.
 if (defined $opt_D) {
     if ($distributor eq 'Debian') {
-	unless ($opt_D =~ /^(unstable|(stable|testing)(-security)?|oldstable-security|experimental|UNRELEASED|(sarge|etch)(-volatile|-backports)?)$/) {
-	    warn "$progname warning: Recognised distributions are: unstable, testing, stable,\nexperimental, UNRELEASED, {sarge,etch}-{volatile,backports} and {testing,stable,oldstable}-security.\nUsing your request anyway.\n";
- 	    $warnings++ if not $opt_force_dist;
+	unless ($opt_D =~ /^(unstable|((old)?stable|testing)(-security)?|experimental|UNRELEASED|(sarge|etch)(-volatile|-backports)?|((oldstable|testing)-)?proposed-updates)$/) {
+	    warn "$progname warning: Recognised distributions are: unstable, testing, stable,\noldstable, experimental, UNRELEASED, {sarge,etch}-{volatile,backports},\n{oldstable-,testing-,}proposed-updates and {testing,stable,oldstable}-security.\nUsing your request anyway.\n";
+	    $warnings++ if not $opt_force_dist;
 	}
     } elsif ($distributor eq 'Ubuntu') {
-	unless ($opt_D =~ /^((warty|hoary|breezy|dapper|edgy|feisty|gutsy|hardy|intrepid)(-updates|-security|-proposed)?|UNRELEASED)$/) {
-	    warn "$progname warning: Recognised distributions are:\n{warty,hoary,breezy,dapper,edgy,feisty,gutsy,hardy,intrepid}{,-updates,-security,-proposed} and UNRELEASED.\nUsing your request anyway.\n";
- 	    $warnings++ if not $opt_force_dist;
+	unless ($opt_D =~ /^((warty|hoary|breezy|dapper|edgy|feisty|gutsy|hardy|intrepid|jaunty)(-updates|-security|-proposed)?|UNRELEASED)$/) {
+	    warn "$progname warning: Recognised distributions are:\n{warty,hoary,breezy,dapper,edgy,feisty,gutsy,hardy,intrepid,jaunty}{,-updates,-security,-proposed} and UNRELEASED.\nUsing your request anyway.\n";
+	    $warnings++ if not $opt_force_dist;
 	}
     } else {
 	# Unknown distributor, skip check
@@ -376,8 +427,9 @@ fatal "--package cannot be used when creating a NEWS file"
     if $opt_package && $opt_news;
 
 if ($opt_create) {
-    if ($opt_a || $opt_i || $opt_e || $opt_r || $opt_b || $opt_n || $opt_security || $opt_qa || $opt_bpo) {
-	warn "$progname warning: ignoring -a/-i/-e/-r/-b/-n/-s/--qa/--bpo options with --create\n";
+    if ($opt_a || $opt_i || $opt_e || $opt_r || $opt_b || $opt_n || $opt_bn ||
+	    $opt_qa || $opt_s || $opt_bpo || $opt_l || $opt_allow_lower) {
+	warn "$progname warning: ignoring -a/-i/-e/-r/-b/--allow-lower-version/-n/--bin-nmu/-q/--qa/-s/--bpo/-l options with --create\n";
 	$warnings++;
     }
     if ($opt_package && $opt_d) {
@@ -487,14 +539,13 @@ if (! $opt_create || ($opt_create && $opt_news)) {
     fatal "No distribution in changelog!"
 	unless exists $changelog{'Distribution'};
     if ($distributor eq 'Ubuntu') {
-	# In Ubuntu uploads should go to hardy
-	$DISTRIBUTION = 'hardy';
+	# In Ubuntu uploads should go to intrepid
+	$DISTRIBUTION = 'intrepid';
     } else {
 	$DISTRIBUTION=$changelog{'Distribution'};
     }
     fatal "No changes in changelog!"
 	unless exists $changelog{'Changes'};
-    $CHANGES=$changelog{'Changes'};
 
     # Find the current package version
     if ($opt_news) {
@@ -649,66 +700,77 @@ if (! $opt_m) {
 
 #####
 
+if ($opt_auto_nmu eq 'yes' and ! $opt_v and ! $opt_l and ! $opt_s and 
+    ! $opt_qa and ! $opt_bpo and ! $opt_bn and ! $opt_n and ! $opt_c and
+    ! $opt_create and ! $opt_a_passed and ! $opt_r and
+    $distributor ne 'Ubuntu') {
+
+    if (-f 'debian/control') {
+	if (have_lpdc()) {
+	    my $parser = new Parse::DebControl;
+	    my $deb822 = $parser->parse_file('debian/control', {stripComments => 'true'});
+	    my $uploader = decode_utf8($deb822->[0]->{'Uploaders'}) || '';
+	    my $maintainer = decode_utf8($deb822->[0]->{'Maintainer'});
+	    my @uploaders = split(/,\s+/, $uploader);
+
+	    my $packager = "$MAINTAINER <$EMAIL>";
+
+	    if (! grep { $_ eq $packager } ($maintainer, @uploaders) and
+		$packager ne $changelog{'Maintainer'}) {
+		$opt_n=1;
+		$opt_a=0;
+	    }
+	} else {
+	    warn "$progname: skipping automatic NMU detection: $lpdc_broken\n";
+	}
+    } else {
+	fatal "Missing file debian/control";
+    }
+}
+#####
+
 # Do we need to generate "closes" entries?
 
 my @closes_text = ();
 my $initial_release = 0;
 if (@closes and $opt_query) { # and we have to query the BTS
-    if (system('command -v wget >/dev/null 2>&1') >> 8 != 0) {
-	warn "$progname warning: wget not installed, so cannot query the bug-tracking system\n";
+    if (!Devscripts::Debbugs::have_soap) {
+	warn "$progname warning: libsoap-lite-perl not installed, so cannot query the bug-tracking system\n";
 	$opt_query=0;
 	$warnings++;
 	# This will now go and execute the "if (@closes and ! $opt_query)" code
     }
     else
     {
-	my %bugs;
-	my $lastbug;
-
-	my $uripkg = uri_escape($PACKAGE);
-	my $bugs = `wget -q -O - 'http://bugs.debian.org/cgi-bin/pkgreport.cgi?src=$uripkg'`;
-	if ($? >> 8 != 0) {
-	    warn "$progname warning: wget failed, so cannot query the bug-tracking system\n";
+	my $bugs = Devscripts::Debbugs::select( "src:" . $PACKAGE );
+	my $statuses = Devscripts::Debbugs::status(
+	    map {[bug => $_, indicatesource => 1]} @{$bugs} );
+	if (not defined $bugs) {
+	    warn "$progname warning: Debbugs::status failed, so cannot query the bug-tracking system\n";
 	    $opt_query=0;
 	    $warnings++;
 	    # This will now go and execute the "if (@closes and ! $opt_query)" code
 	}
-
-	foreach (split /\n/, $bugs) {
-	    if (m%<a(?: class=\".*?\")? href=\"bugreport.cgi\?bug=([0-9]*).*?>\#\1: (.*?)</a>%i) {
-		$bugs{$1} = [$2];
-		$lastbug=$1;
-	    }
-	    elsif (defined $lastbug and
-		   m%<a(?: class=\".*?\")? href=\"pkgreport.cgi\?(?:[^\"]*?;)(?:(pkg|package))=([a-z0-9\+\-\.]*)%i) {
-		push @{$bugs{$lastbug}}, $1
-		    if exists $bugs{$lastbug};
-		$lastbug = undef;
-	    }
-	}
-
 	foreach my $close (@closes) {
-	    if (exists $bugs{$close}) {
-		my ($title,$pkg) = @{$bugs{$close}};
+	    if (exists $statuses->{$close}) {
+		my $title = $statuses->{$close}->{subject};
+		my $pkg = $statuses->{$close}->{package};
 		$title =~ s/^($pkg|$PACKAGE): //;
-		$title =~ s/&quot;/\"/g;
-		$title =~ s/&lt;/</g;
-		$title =~ s/&gt;/>/g;
-		$title =~ s/&amp;/&/g;
 		push @closes_text, "$title (Closes: \#$close)\n";
 	    }
 	    else { # not our package, or wnpp
-		my $bug = `wget -q -O - 'http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=$close'`;
-		if ($? >> 8 != 0) {
+		my $bug = Devscripts::Debbugs::status(
+		    [bug => $close, indicatesource => 1] );
+		if (not defined $bug) {
 		    warn "$progname warning: unknown bug \#$close does not belong to $PACKAGE,\n  disabling closing changelog entry\n";
 		    $warnings++;
 		    push @closes_text, "Closes?? \#$close: UNKNOWN BUG IN WRONG PACKAGE!!\n";
 		} else {
-		    my ($bugtitle) = ($bug =~ m%<TITLE>.*?\#$close - (.*?)</TITLE>%);
+		    my $bugtitle = $bug->{$close}->{subject};
 		    $bugtitle ||= '';
-		    my ($bugpkg) = ($bug =~ m%<a(?: class=\".*?\")? href=\"pkgreport.cgi\?(?:[^\"]*?;)(?:(pkg|package))=([a-z0-9\+\-\.]*)%i);
+		    my $bugpkg = $bug->{$close}->{package};
 		    $bugpkg ||= '?';
-		    my ($bugsrcpkg) = ($bug =~ m%<a(?: class=\".*?\")? href=\"pkgreport.cgi\?src=([a-z0-9\+\-\.]*)%i);
+		    my $bugsrcpkg = $bug->{$close}->{source};
 		    $bugsrcpkg ||= '?';
 		    if ($bugsrcpkg eq $PACKAGE) {
 			warn "$progname warning: bug \#$close appears to be already archived,\n  disabling closing changelog entry\n";
@@ -764,9 +826,18 @@ if (@ARGV and ! $TEXT) {
 my $date_cmd = ($opt_tz ? "TZ=$opt_tz " : "") . "date -R";
 chomp(my $DATE=`$date_cmd`);
 
+if ($opt_news && !$opt_i && !$opt_a) {
+    if ($VERSION eq $changelog{'Version'}) {
+	$opt_a = 1;
+    } else {
+	$opt_i = 1;
+    }
+}
+
 # Are we going to have to figure things out for ourselves?
 if (! $opt_i && ! $opt_v && ! $opt_d && ! $opt_a && ! $opt_e && ! $opt_r &&
-    ! $opt_create) {
+    ! $opt_n && ! $opt_bn && ! $opt_qa && ! $opt_s && ! $opt_bpo &&
+    ! $opt_l && ! $opt_create) {
     # Yes, we are
     if ($opt_release_heuristic eq 'log') {
 	my @UPFILES = glob("../$PACKAGE\_$SVERSION\_*.upload");
@@ -799,7 +870,9 @@ if (! $opt_i && ! $opt_v && ! $opt_d && ! $opt_a && ! $opt_e && ! $opt_r &&
 	if ($changelog{'Distribution'} eq 'UNRELEASED') {
 		$opt_a = 1;
 	}
-	else {
+	elsif ($EMPTY_TEXT==1) {
+		$opt_a = 1;
+	} else {
 		$opt_i = 1;
 	}
     }
@@ -811,6 +884,22 @@ if (! $opt_i && ! $opt_v && ! $opt_d && ! $opt_a && ! $opt_e && ! $opt_r &&
 # Open in anticipation....
 unless ($opt_create) {
     open S, $changelog_path or fatal "Cannot open existing $changelog_path: $!";
+
+    # Read the first stanza from the changelog file
+    # We do this directly rather than reusing $changelog{'Changes'}
+    # so that we have the verbatim changes rather than a (albeit very
+    # slightly) reformatted version. See Debian bug #452806
+
+    while(<S>) {
+	last if /^ --/;
+
+	$CHANGES .= $_;
+    }
+
+    chomp $CHANGES;
+
+    # Reset file pointer
+    seek(S, 0, 0);
 }
 open O, ">$changelog_path.dch"
     or fatal "Cannot write to temporary file: $!";
@@ -821,10 +910,14 @@ select((select(O), $^L = "")[0]);
 my $tmpchk=1;
 my ($NEW_VERSION, $NEW_SVERSION, $NEW_UVERSION);
 my $line;
+my $optionsok=0;
+my $merge=0;
 
-if (($opt_i || $opt_n || $opt_security || $opt_qa || $opt_bpo || $opt_v || $opt_d ||
+if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $opt_v || $opt_d ||
     ($opt_news && $VERSION ne $changelog{'Version'})) && ! $opt_create) {
 
+    $optionsok=1;
+
     # Check that a given explicit version number is sensible.
     if ($opt_v || $opt_d) {
 	if($opt_v) {
@@ -866,7 +959,7 @@ if (($opt_i || $opt_n || $opt_security || $opt_qa || $opt_bpo || $opt_v || $opt_
 
 	if (system("dpkg --compare-versions $VERSION lt $NEW_VERSION" .
 		  " 2>/dev/null 1>&2")) {
-	    if ($opt_b) {
+	    if ($opt_b or ($opt_allow_lower and $NEW_VERSION =~ /$opt_allow_lower/)) {
 		warn "$progname warning: new version ($NEW_VERSION) is less than\n" .
 		    "the current version number ($VERSION).\n";
 	    } else {
@@ -889,38 +982,48 @@ if (($opt_i || $opt_n || $opt_security || $opt_qa || $opt_bpo || $opt_v || $opt_
 	if ($VERSION =~ /(.*?)([a-yA-Y][a-zA-Z]*|\d+)$/i) {
 	    my $end=$2;
 	    my $start=$1;
-	    if ($distributor ne 'Ubuntu' || $start =~ /(build|ubuntu|prop(osed)?)(\d+\.)*$/ || $opt_U) {
 	    # If it's not already an NMU make it so
 	    # otherwise we can be safe if we behave like dch -i
-	    if (($opt_n || $opt_security) and (not $start =~ /\.$/ or $VERSION eq $UVERSION)) {
+
+	    if ($distributor eq 'Ubuntu' and $start !~ /(build|ubuntu|prop(osed)?)(\d+\.)*$/ and not $opt_U) {
+		if ($start =~ /~ppa([0-9]*)$/) {
+		    $end++;
+		} else {
+		    $end .= "ubuntu1";
+		}
+	    } elsif (($opt_n or $opt_s) and (
+		($VERSION eq $UVERSION and not $start =~ /\+nmu/)
+		or ($VERSION ne $UVERSION and not $start =~ /\.$/))) {
+
 		if ($VERSION eq $UVERSION) {
 		    # First NMU of a Debian native package
-		    $end .= "-0.1";
+		    $end .= "+nmu1";
 		} else {
 	    	    $end += 0.1;
 		}
+            } elsif ($opt_bn and not $start =~ /\+b/) {
+                $end .= "+b1";
 	    } elsif ($opt_qa and $start =~/(.*?)-(\d+)\.$/) {
-		    # Drop NMU revision when doing a QA upload
-		    my $upstream_version = $1;
-		    my $debian_revision = $2;
-		    $debian_revision++;
-		    $start = "$upstream_version-$debian_revision";
-		    $end = "";
+                # Drop NMU revision when doing a QA upload
+                my $upstream_version = $1;
+                my $debian_revision = $2;
+                $debian_revision++;
+                $start = "$upstream_version-$debian_revision";
+                $end = "";
 	    } elsif ($opt_bpo and not $start =~ /~bpo\.$/) {
 		# If it's not already a backport make it so
 		# otherwise we can be safe if we behave like dch -i
 		$end .= "~bpo40+1";
+	    } elsif ($opt_l and not $start =~ /\Q$opt_l\E/) {
+		# If it's not already a local package make it so
+		# otherwise we can be safe if we behave like dch -i
+		$end .= $opt_l."1";
 	    } elsif (!$opt_news) {
 		# Don't bump the version of a NEWS file in this case as we're
 		# using the version from the changelog
 		$end++;
 	    }
-	    $DISTRIBUTION=$changelog{'Distribution'} if $start =~ /prop(osed)?(\d+\.)*$/;
-	    } elsif ($start =~ /~ppa([0-9]*)$/) {
-		$end++
-	    } else {
-		$end .= "ubuntu1";
-	    }
+	    $DISTRIBUTION = $changelog{'Distribution'} if $start =~ /prop(osed)?(\d+\.)*$/;
 	    $NEW_VERSION = "$start$end";
 	    ($NEW_SVERSION=$NEW_VERSION) =~ s/^\d+://;
 	    ($NEW_UVERSION=$NEW_SVERSION) =~ s/-[^-]*$//;
@@ -929,50 +1032,72 @@ if (($opt_i || $opt_n || $opt_security || $opt_qa || $opt_bpo || $opt_v || $opt_
 	}
     }
 
+    if ($NEW_VERSION eq $NEW_UVERSION and $VERSION ne $UVERSION) {
+	warn "$progname warning: New package version is Debian native whilst previous version was not\n";
+    } elsif ($NEW_VERSION ne $NEW_UVERSION and $VERSION eq $UVERSION) {
+	warn "$progname warning: Previous package version was Debian native whilst new version is not\n"
+	    unless $opt_n or $opt_s;
+    }
+
     if ($opt_bpo) {
 	$bpo_dist='etch-backports';
     }
     my $distribution = $opt_D || $bpo_dist || (($opt_release_heuristic eq 'changelog') ? "UNRELEASED" : $DISTRIBUTION);
     
     my $urgency = $opt_u || 'low';
-    print O "$PACKAGE ($NEW_VERSION) $distribution; urgency=$urgency\n\n";
 
-    if ($opt_n && ! $opt_news) {
-	print O "  * Non-maintainer upload.\n";
-	$line = 1;
-    } elsif ($opt_qa && ! $opt_news) {
-	print O "  * QA upload.\n";
-	$line = 1;
-    } elsif ($opt_bpo && ! $opt_news) {
-	print O "  * Rebuild for Etch backports.\n";
-        $line = 1;
-    }
-    if (@closes_text or $TEXT) {
-	foreach (@closes_text) { format_line($_, 1); }
-	if (length $TEXT) { format_line($TEXT, 1); }
-    } elsif ($opt_news) {
-	print O "  \n";
-    } elsif ($opt_security) {
-	print O "  * SECURITY UPDATE: \n";
-	print O "  * References\n";
-	$line += 1;
+    if (($opt_v or $opt_i or $opt_l) and $opt_release_heuristic eq "changelog" and
+	$changelog{'Distribution'} eq "UNRELEASED" and
+	$distribution eq "UNRELEASED") {
+
+	$merge = 1;
     } else {
-	print O "  * \n";
-    }
-    $line += 3;
-    print O "\n -- $MAINTAINER <$EMAIL>  $DATE\n\n";
+	print O "$PACKAGE ($NEW_VERSION) $distribution; urgency=$urgency\n\n";
+	if ($opt_n && ! $opt_news) {
+            print O "  * Non-maintainer upload.\n";
+	    $line = 1;
+	} elsif ($opt_bn && ! $opt_news) {
+            my $arch = qx/dpkg-architecture -qDEB_BUILD_ARCH/; chomp ($arch);
+            print O "  * Binary-only non-maintainer upload for $arch; no source changes.\n";
+            $line = 1;
+	} elsif ($opt_qa && ! $opt_news) {
+            print O "  * QA upload.\n";
+            $line = 1;
+	} elsif ($opt_s && ! $opt_news) {
+	    if ($distributor eq 'Ubuntu') {
+		print O "  * SECURITY UPDATE:\n";
+		print O "  * References\n";
+	    } else {
+		print O "  * Non-maintainer upload by the Security Team.\n";
+	    }
+            $line = 1;
+	} elsif ($opt_bpo && ! $opt_news) {
+            print O "  * Rebuild for etch-backports.\n";
+            $line = 1;
+	}
+	if (@closes_text or $TEXT) {
+            foreach (@closes_text) { format_line($_, 1); }
+            if (length $TEXT) { format_line($TEXT, 1); }
+	} elsif ($opt_news) {
+            print O "  \n";
+	} else {
+            print O "  * \n";
+	}
+	$line += 3;
+	print O "\n -- $MAINTAINER <$EMAIL>  $DATE\n\n";
 
-    # Copy the old changelog file to the new one
-    local $/ = undef;
-    print O <S>;
+	# Copy the old changelog file to the new one
+	local $/ = undef;
+	print O <S>;
+    }
 }
-elsif (($opt_r || $opt_a) && ! $opt_create) {
+if (($opt_r || $opt_a || $merge) && ! $opt_create) {
     # This means we just have to generate a new * entry in changelog
     # and if a multi-developer changelog is detected, add developer names.
     
-    $NEW_VERSION=$VERSION;
-    $NEW_SVERSION=$SVERSION;
-    $NEW_UVERSION=$UVERSION;
+    $NEW_VERSION=$VERSION unless $NEW_VERSION;
+    $NEW_SVERSION=$SVERSION unless $NEW_SVERSION;
+    $NEW_UVERSION=$UVERSION unless $NEW_UVERSION;
 
     # Read and discard maintainer line, see who made the
     # last entry, and determine whether there are existing
@@ -1050,11 +1175,13 @@ elsif (($opt_r || $opt_a) && ! $opt_create) {
 	my $urgency = $2;
 	$distribution =~ s/^\s+//;
 	if ($opt_r) {
-	    # Change the distribution from UNRELEASED for release.
+	    # Change the distribution from UNRELEASED for release
 	    if ($distributor eq 'Ubuntu') {
-		$distribution = $opt_D || "hardy";
-	    } else {
+		$distribution = $opt_D || "intrepid";
+	    } elsif ($distribution eq "UNRELEASED") {
 		$distribution = $opt_D || $lastdist || "unstable";
+	    } elsif ($opt_D) {
+		warn "$progname warning: ignoring distribution passed to --release as changelog has already been released\n";
 	    }
 	    # Set the start-line to 1, as we don't know what they want to edit
 	    $line=1;
@@ -1062,7 +1189,7 @@ elsif (($opt_r || $opt_a) && ! $opt_create) {
 	    $distribution = $opt_D if $opt_D;
 	}
 	$urgency = $opt_u if $opt_u;
-	$CHANGES =~ s/^(\w[-+0-9a-z.]* \([^\(\) \t]+\))(?:\s+[-+0-9a-z.]+)+\;\s+urgency=\w+/$1 $distribution; urgency=$urgency/i;
+	$CHANGES =~ s/^(\w[-+0-9a-z.]* \([^\(\) \t]+\))(?:\s+[-+0-9a-z.]+)+\;\s+urgency=\w+/$PACKAGE ($NEW_VERSION) $distribution; urgency=$urgency/i;
     } else {
 	warn "$progname: couldn't parse first changelog line, not touching it\n";
 	$warnings++;
@@ -1151,7 +1278,8 @@ elsif ($opt_e && ! $opt_create) {
     $line=0;
 }
 elsif ($opt_create) {
-    if (! $initial_release and ! $opt_news) {
+    if (! $initial_release and ! $opt_news and ! $opt_empty and
+	! $TEXT and ! $EMPTY_TEXT) {
 	push @closes_text, "Initial release. (Closes: \#XXXXXX)\n";
     }
 
@@ -1163,6 +1291,8 @@ elsif ($opt_create) {
 	if (length $TEXT) { format_line($TEXT, 1); }
     } elsif ($opt_news) {
 	print O "  \n";
+    } elsif ($opt_empty) {
+	# Do nothing, but skip the empty entry
     } else { # this can't happen, but anyway...
 	print O "  * \n";
     }
@@ -1171,7 +1301,7 @@ elsif ($opt_create) {
 
     $line = 1;
 }
-else {
+elsif (!$optionsok) {
     fatal "Unknown changelog processing command line options - help!";
 }
 
@@ -1190,12 +1320,14 @@ if ($warnings) {
 }
 
 # Now Run the Editor; always run if doing "closes" to give a chance to check
-if ((!$TEXT and !$EMPTY_TEXT) or @closes_text or $opt_create) {
+if ((!$TEXT and !$EMPTY_TEXT and ! ($opt_create and $opt_empty)) or @closes_text or
+    ($opt_create and ! ($PACKAGE ne 'PACKAGE' and $VERSION ne 'VERSION'))) {
+
     my $mtime = (stat("$changelog_path.dch"))[9];
     defined $mtime or fatal
 	"Error getting modification time of temporary $changelog_path: $!";
 
-    $line-- if ($opt_security);
+    $line-- if $opt_s and $distributor eq 'Ubuntu';
     system("sensible-editor +$line $changelog_path.dch") == 0 or
 	fatal "Error editing $changelog_path";
 
@@ -1271,6 +1403,13 @@ sub format_line {
     $CHGLINE=shift;
     my $newentry=shift;
 
+    # Work around the fact that write() with formats
+    # seems to assume that character == byte
+    # See http://rt.perl.org/rt3/Public/Bug/Display.html?id=33832
+    # and Debian bug #473769
+    my $count = () = $CHGLINE =~ /[^\x00-\x7F]/mg;
+    $CHGLINE .= " " x $count;
+
     print O "\n" if $opt_news && ! ($newentry || $linecount);
     $linecount++;
     my $f=select(O);
