@@ -27,6 +27,7 @@ use strict;
 use Cwd;
 use File::Basename;
 use File::Copy;
+use File::Temp qw/tempdir/;
 use filetest 'access';
 use Getopt::Long;
 use lib '/usr/share/devscripts';
@@ -92,7 +93,8 @@ Options:
     --symlink      Make an orig.tar.gz symlink to downloaded file (default)
     --rename       Rename to orig.tar.gz instead of symlinking
                    (Both will use orig.tar.bz2 if appropriate)
-    --repack       Repack downloaded archives from orig.tar.bz2 to orig.tar.gz
+    --repack       Repack downloaded archives from orig.tar.bz2 or orig.zip to
+                   orig.tar.gz
                    (does nothing if downloaded archive orig.tar.gz)
     --no-symlink   Don\'t make symlink or rename
     --verbose      Give verbose output
@@ -113,6 +115,10 @@ Options:
     --upstream-version VERSION
                    Specify the current upstream version in use rather than
                    parsing debian/changelog to determine this
+    --download-version VERSION
+                   Specify the version which the upstream release must
+                   match in order to be considered, rather than using the
+                   release with the highest version
     --package PACKAGE
                    Specify the package name rather than examining
                    debian/changelog; must use --upstream-version and
@@ -153,9 +159,10 @@ our $passive = 'default';
 
 my $destdir = "..";
 my $download = 1;
+my $download_version;
 my $force_download = 0;
 my $report = 0; # report even on up-to-date packages?
-my $repack = 0; # repack .tar.bz2 to .tar.gz
+my $repack = 0; # repack .tar.bz2 or .zip to .tar.gz
 my $symlink = 'symlink';
 my $verbose = 0;
 my $check_dirname_level = 1;
@@ -240,7 +247,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $check_dirname_level = $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'};
     $check_dirname_regex = $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_REGEX'};
     $user_agent_string = $config_vars{'USCAN_USER_AGENT'}
-	if defined $config_vars{'USCAN_USER_AGENT'};
+	if $config_vars{'USCAN_USER_AGENT'};
 }
 
 # Now read the command line arguments
@@ -249,12 +256,14 @@ my ($opt_h, $opt_v, $opt_destdir, $opt_download, $opt_force_download,
     $opt_report, $opt_passive, $opt_symlink, $opt_repack);
 my ($opt_verbose, $opt_ignore, $opt_level, $opt_regex, $opt_noconf);
 my ($opt_package, $opt_uversion, $opt_watchfile, $opt_dehs, $opt_timeout);
+my $opt_download_version;
 my $opt_user_agent;
 
 GetOptions("help" => \$opt_h,
 	   "version" => \$opt_v,
 	   "destdir=s" => \$opt_destdir,
 	   "download!" => \$opt_download,
+	   "download-version=s" => \$opt_download_version,
 	   "force-download" => \$opt_force_download,
 	   "report" => sub { $opt_download = 0; },
 	   "report-status" => sub { $opt_download = 0; $opt_report = 1; },
@@ -299,6 +308,7 @@ $symlink = $opt_symlink if defined $opt_symlink;
 $verbose = $opt_verbose if defined $opt_verbose;
 $dehs = $opt_dehs if defined $opt_dehs;
 $user_agent_string = $opt_user_agent if defined $opt_user_agent;
+$download_version = $opt_download_version if defined $opt_download_version;
 if ($dehs) {
     $SIG{'__WARN__'} = \&dehs_warn;
     $SIG{'__DIE__'} = \&dehs_die;
@@ -319,8 +329,8 @@ if (defined $opt_level) {
 $check_dirname_regex = $opt_regex if defined $opt_regex;
 
 if (defined $opt_package) {
-    die "$progname: --package requires the use of --upstream-version and --watchfile\nas well; run $progname --help for more details\n"
-	unless defined $opt_uversion and defined $opt_watchfile;
+    die "$progname: --package requires the use of --watchfile\nas well; run $progname --help for more details\n"
+	unless defined $opt_watchfile;
     $download = -$download unless defined $opt_download;
 }
 
@@ -508,6 +518,8 @@ for my $dir (@dirs) {
     }
 }
 
+warn "$progname: no watch file found\n" if (@debdirs == 0 and $report);
+
 # Was there a --uversion option?
 if (defined $opt_uversion) {
     if (@debdirs == 1) {
@@ -735,6 +747,11 @@ sub process_watchline ($$$$$$)
 	    (undef, $lastversion, $action) = split ' ', $line, 3;
 	}
 
+	if ((!$lastversion or $lastversion eq 'debian') and not defined $pkg_version) {
+	    warn "$progname warning: Unable to determine current version\n  in $watchfile, skipping:\n  $line\n";
+	    return 1;
+	}
+
 	# Check all's OK
 	if ($filepattern !~ /\(.*\)/) {
 	    warn "$progname warning: Filename pattern missing version delimiters ()\n  in $watchfile, skipping:\n  $line\n";
@@ -842,8 +859,19 @@ sub process_watchline ($$$$$$)
 		print "-- Found the following matching hrefs:\n";
 		foreach my $href (@hrefs) { print "     $$href[1]\n"; }
 	    }
-	    @hrefs = Devscripts::Versort::versort(@hrefs);
-	    ($newversion, $newfile) = @{$hrefs[0]};
+	    if (defined $download_version) {
+		my @vhrefs = grep { $$_[0] eq $download_version } @hrefs;
+		if (@vhrefs) {
+		    ($newversion, $newfile) = @{$vhrefs[0]};
+		} else {
+		    warn "$progname warning: In $watchfile no matching hrefs for version $download_version"
+			. " in watch line\n  $line\n";
+		    return 1;
+		}
+	    } else {
+		@hrefs = Devscripts::Versort::versort(@hrefs);
+		($newversion, $newfile) = @{$hrefs[0]};
+	    }
 	} else {
 	    warn "$progname warning: In $watchfile,\n  no matching hrefs for watch line\n  $line\n";
 	    return 1;
@@ -926,8 +954,19 @@ sub process_watchline ($$$$$$)
 		print "-- Found the following matching files:\n";
 		foreach my $file (@files) { print "     $$file[1]\n"; }
 	    }
-	    @files = Devscripts::Versort::versort(@files);
-	    ($newversion, $newfile) = @{$files[0]};
+	    if (defined $download_version) {
+		my @vfiles = grep { $$_[0] eq $download_version } @files;
+		if (@vfiles) {
+		    ($newversion, $newfile) = @{$vfiles[0]};
+		} else {
+		    warn "$progname warning: In $watchfile no matching files for version $download_version"
+			. " in watch line\n  $line\n";
+		    return 1;
+		}
+	    } else {
+		@files = Devscripts::Versort::versort(@files);
+		($newversion, $newfile) = @{$files[0]};
+	    }
 	} else {
 	    warn "$progname warning: In $watchfile no matching files for watch line\n  $line\n";
 	    return 1;
@@ -978,7 +1017,12 @@ EOF
 	}
     }
     if (! $lastversion or $lastversion eq 'debian') {
-	$lastversion=$pkg_version;
+	if (defined $pkg_version) {
+	    $lastversion=$pkg_version;
+	} else {
+	    warn "$progname warning: Unable to determine current version\n  in $watchfile, skipping:\n  $line\n";
+	    return 1;
+	}
     }
     # And mangle it if requested
     my $mangled_lastversion = $lastversion;
@@ -1063,18 +1107,24 @@ EOF
 
     # We use dpkg's rules to determine whether our current version
     # is newer or older than the remote version.
-    if (system("dpkg", "--compare-versions", "'$mangled_lastversion'", "gt", "'$newversion'") == 0) {
-        if ($verbose) {
-	    print " => remote site does not even have current version\n";
-	} elsif ($dehs) {
-	    $dehs_tags{'status'} = "Debian version newer than remote site";
+    if (!defined $download_version) {
+        if (system("dpkg", "--compare-versions", "'$mangled_lastversion'", "gt", "'$newversion'") == 0) {
+	    if ($verbose) {
+		print " => remote site does not even have current version\n";
+	    } elsif ($dehs) {
+		$dehs_tags{'status'} = "Debian version newer than remote site";
+	    } else {
+		print "$pkg: remote site does not even have current version\n";
+	    }
+	    return 0;
 	} else {
-	    print "$pkg: remote site does not even have current version\n";
+	    # There's a newer upstream version available, which may already
+	    # be on our system or may not be
+	    $found++;
 	}
-        return 0;
     } else {
-	# There's a newer upstream version available, which may already
-	# be on our system or may not be
+	# Flag that we found a newer upstream version, so that the exit status
+	# is set correctly
 	$found++;
     }
 
@@ -1181,6 +1231,20 @@ EOF
 	$newfile_base = $newfile_base_gz;
     }
 
+    if ($repack and $newfile_base =~ /^(.*)\.zip$/) {
+	print "-- Repacking from zip to .tar.gz\n" if $verbose;
+
+	system('command -v unzip >/dev/null 2>&1') >> 8 == 0
+	  or die("unzip binary not found. You need to install the package unzip to be able to repack .zip upstream archives.\n");
+
+	my $newfile_base_gz = "$1.tar.gz";
+	my $tempdir = tempdir ( "uscanXXXX", TMPDIR => 1, CLEANUP => 1 );
+	system("unzip -q -d $tempdir $destdir/$newfile_base; tar -C $tempdir -czf $destdir/$newfile_base_gz .") == 0 
+	  or die("Repacking from zip to tar.gz failed\n");
+	unlink "$destdir/$newfile_base";
+	$newfile_base = $newfile_base_gz;
+    }
+
     if ($newfile_base =~ /\.(tar\.gz|tgz|tar\.bz2|tbz2?)$/) {
 	my $filetype = `file $destdir/$newfile_base`;
 	$filetype =~ s%^\.\./\Q$newfile_base\E: %%;
